/******************************************************************************
File:	 	cert.c
Description: 	Handles the internal machinery of creating certificates and 
		performing certificate steps.

Comments:    	TODO:
		
	
Created:	Z.Bukovac Jan 2012
*******************************************************************************/
#include <stdlib.h>
#include <stdbool.h>
#include "graph.h"
#include "cert.h"

typedef struct operationListNode *operationListNode;

enum operationPrecedence {DELADDINV = 0, ADDDELINV = 1, DELOPT = 2, ADDOPT = 3, FAST = 4};
enum edgePrecedence {STARTSMALL = 0, STARTBIG = 1};

const enum operationPrecedence defaultPrecedence = FAST;
const enum edgePrecedence defaultEdgePrecedence = STARTSMALL;

cert newCert(graph initalGraph);
line newLine(void);
operation newOperation(cert inpCert, enum operationType type, int g1, int g2, int u, int v);
operationList newOperationList(void);
operationListNode newOperationListNode(operation op);
void addOpToList(operationList list, operation op);
line initFirstLine(graph inpGraph);
int getInitialGraphOrder(cert inpCert);
bool isInitialGraph(cert inpCert, graph testGraph);
bool endGraphIsInitialGraph(cert inpCert);
bool isLastGraph(cert inpCert, graph testGraph);
void printCert(cert inpCert);
void printCertToFile(cert inpCert, FILE *outFile);
void printLine(line inpLine);
void printLineToFile(line inpLine, FILE *outFile);
void printCertGraphs(cert inpCert);
void printCertGraphsToFile(cert inpCert, FILE *outFile);
void printOperation(operation op);
void addNewLine(cert inpCert, line inpLine);
void removeLastLine(cert inpCert);
int freeLine(line inpLine);
void freeList(operationList inpList);
line getCurrentLine(cert inpCert);
void performOp(operation op);
void deleteCont(cert inpCert, int graphIndex, int u, int v);
void addIdent(cert inpCert, int graphIndex, int u, int v);
void deleteContRev1(cert inpCert, int graphIndex1, int graphIndex2, int u, int v);
void deleteContRev2(cert inpCert, int graphIndex1, int graphIndex2, int u, int v);
void addIdentRev(cert inpCert, int graphIndex1, int graphIndex2, int u, int v);
bool willCancel(line inpLine, int graphIndex1, int graphIndex2);
operationList getCancelOperationList(cert inpCert);
void printCancelableGraphPairs(cert inpCert);
void cancelGraphs(cert inpCert, int graphIndex1, int graphIndex2);
int willDC2Reverse(line inpLine, int graphIndex1, int graphIndex2);
int willAIReverse(line inpLine, int graphIndex1, int graphIndex2);
void printCurrentGraphList(cert inpCert);
int getCurrentLineLength(cert inpCert);
int getLength(cert inpCert);
graph getGraph(cert inpCert, int index);
int getIdAtIndex(cert inpCert, int index);
bool canDeleteCont(cert inpCert, int graphIndex, int u, int v);
bool canAddIdent(cert inpCert, int graphIndex, int u, int v);
operationList getDC1RevOperationList(cert inpCert);
operationList getDC2RevOperationList(cert inpCert);
operationList getAIRevOperationList(cert inpCert);
operationList getDelContOperationList(cert inpCert);
operationList getAddIdentOperationList(cert inpCert);
operationList getOperationList(cert inpCert, enum operationType optype);
bool listIsEmpty(operationList list);
int getListLength(operationList list);
operationList getAllOperations(cert inpCert);
operationList getAllCompressingOperations(cert inpCert);
void appendList(operationList list1, operationList list2);
operation getOpFromList(operationList list, int opIndex);
void printOperationList(operationList list);
void runComp(cert inpCert);
void certTest();
void newTest();
operationList buildOperationList(cert inpCert, enum operationType * types, int size);



/******************************************************************************
**********************************BEGIN STRUCTS********************************
*******************************************************************************/

struct operation
{
	cert cert;
	int graphIndex1;
	int graphIndex2;
	int u;
	int v;
	enum operationType optype;
};

struct operationList
{
	operationListNode head;
	operationListNode tail;
	int length;
};

struct operationListNode
{
	operation op;
	operationListNode next;
};

struct cert
{
	line head;
	line tail;
	int length;
	graph graph;
	int graphCounter;
};

struct line
{
	int numGraphs;
	graph*graphs;
	int *signs;	
	line next;
	line prev;
	char description[20];
};

/******************************************************************************
***********************************END STRUCTS*********************************
*******************************************************************************/


/******************************************************************************
*********************************BEGIN FUNCTIONS*******************************
*******************************************************************************/

/******************************************************************************
Function:	newCert
Description:	Returns a new cert
Inputs:		graph - the graph that this certificate is for.
Return Value:	cert - 	the new certificate
Side Effects:	Allocates room for and initialises an new cert struct
Notes:		cert (or certificate) structs are essentially an list of lines.
		They also store some other housekkeping data to assist with
		the creation of certificates. Note that while a certificate in
		the computational complexity sense needs to start and end with
		the graphs or factorisation you're looking for, these certs
		guarentee no such completion. A cert struct may very well store
		an actual finished certificate, or it may be partial, and just
		be some lines of algebraic expressions.
Created:	Z.Bukovac Feb 2012
******************************************************************************/
cert newCert(graph initialGraph)
{
	cert cert = NULL;
	cert = malloc(sizeof(* cert));
	cert->length = 0;
	cert->head = initFirstLine(initialGraph);
	cert->tail = cert->head;
	cert->graph = createGraphRef(initialGraph);
	cert->graphCounter = 1;
	return cert;
}

/******************************************************************************
Function:	newLine
Description:	Returns a new line
Inputs:		None
Return Value:	line - 	the new line, containing nothing
Side Effects:	Allocates room for and initialises an new line struct
Notes:		line structs are used to store representations of the algebraic
		expressions in a certificate line. Be aware that the signs pointer
		is not initialised here, and is allocated some space only when
		some actual graphs are to be added to the line, as before then
		we have no idea just how much room will be needed.
Created:	Z.Bukovac Feb 2012
******************************************************************************/
line newLine(void)
{
	line line = NULL;
	line = malloc(sizeof(* line));
	line->graphs = NULL;
	line->next = NULL;
	line->prev = NULL;
	line->numGraphs = 0;
	sprintf(line->description,"Unknown");
	return line;
}

/******************************************************************************
Function:	newOperation	
Description:	Returns a new operation struct
Inputs:		inpCert - 	The certificate on which the operation is to be perfomed
		type	-	The certificates step type
		g1	- 	The index of the first graph involved in the op.
		g2	- 	The index of the second graph involved in the op, if any.
		u, v	-	The edge / non-edge  involved in the op, if any.
Return Value:	op - 	The new operation
Side Effects:	Allocates room for and initialises an new operation struct
Notes:		The operation is always defined on the current last line of the inpCert.
		In some ways, this is less than ideal. Since an operation is performed on a
		line in a certificate, not really on a certificate as a whole, it's a little
		confusing at first. However, it does have the advantage of completely separating
		out the need for the function that calls performOp on an operation struct from
		having any knowledge of the existance of the associated certficate. 
Created:	Z.Bukovac Feb 2012
******************************************************************************/
operation newOperation(cert inpCert, enum operationType type, int g1, int g2, int u, int v)
{
	operation op = NULL;
	op = malloc(sizeof(*op));
	op->cert = inpCert;
	op->graphIndex1 = g1;
	op->graphIndex2 = g2;
	op->v = v;
	op->u = u;
	op->optype = type;
	return op;
}

operationList newOperationList(void)
{
	operationList list = NULL;
	list = malloc(sizeof(*list));
	list->head = NULL;
	list->length = 0;
	return list;
}

operationListNode newOperationListNode(operation op)
{
	operationListNode node = NULL;
	node = malloc(sizeof(*node));
	node->op = op;
	node->next = NULL;
	return node;
}


void addOpToList(operationList list, operation op)
{
	operationListNode node = newOperationListNode(op);
	
	if(list->head == NULL)
	{
		list->head = node;
		list->tail = node;
		//printf("\n\nAdding to head %d",node->op);
	}
	else
	{
		list->tail->next = node;
		list->tail = node;
		//printf("\n\nAdding to tail%d",node->op);
	}
	
	list->length++;
}

line initFirstLine(graph inpGraph)
{
	line line = newLine();
	line->graphs = malloc(sizeof(graph));
	line->signs = malloc(sizeof(int));
	line->graphs[0] = createGraphRef(inpGraph);
	setID(inpGraph, 0);
	line->signs[0] = 1;
	line->numGraphs = 1;
	return line;
}

int getInitialGraphOrder(cert inpCert)
{
	return getSize(inpCert->graph);
}

bool isInitialGraph(cert inpCert, graph testGraph)
{
	return isomorphic(inpCert->graph, testGraph);
}

bool endGraphIsInitialGraph(cert inpCert)
{
	if(inpCert->tail->numGraphs == 1)
	{
		return isomorphic(inpCert->tail->graphs[0], inpCert->graph);
	}
	else return false;
}

bool isLastGraph(cert inpCert, graph testGraph)
{
	if(inpCert->tail->numGraphs == 1)
	{
		return isomorphic(inpCert->tail->graphs[0], testGraph);
	}
	else return false;
}

/******************************************************************************
Function:	printCert
Description:	Prints the given certificate
Inputs:		cert - the certificate to print
Return Value:	None
Side Effects:	None
Created:	Z.Bukovac Feb 2012
******************************************************************************/
void printCert(cert inpCert)
{
	int i;
	line currentLine = inpCert->head->next;
	printf("\nCertificate Length: %d\n", inpCert->length);
	printf("G =");
	for(i = 0; i < inpCert->length; i++) //Note that the first line which contains just the first graph is skipped
	{
		//Print each line in the Certificate
		if(i!=0)	printf("    =");
		printf("\t");
		printLine(currentLine);
		printf("\n");
		currentLine = currentLine->next;
	}
	printf("\n");
}

void printCertToFile(cert inpCert, FILE *outFile)
{
	int i;
	line currentLine = inpCert->head->next;
	fprintf(outFile,"CERT LENGTH: %d\n\n", inpCert->length);
	fprintf(outFile,"G   =");
	for(i = 0; i < inpCert->length; i++) //Note that the first line which contains just the first graph is skipped
	{
		//Print each line in the Certificate
		if(i!=0)	fprintf(outFile,"    =");
		fprintf(outFile,"\t");
		printLineToFile(currentLine, outFile);
		fprintf(outFile,"\n");
		currentLine = currentLine->next;
	}

	fprintf(outFile,"\n");
	printCertSchemaToFile(inpCert,outFile);
	fprintf(outFile,"\n");
}

void printCertSchemaToFile(cert inpCert, FILE *outFile)
{
	int i;
	line currentLine = inpCert->head->next;
	fprintf(outFile,"Schema:{");
	for(i = 0; i < inpCert->length; i++) //Note that the first line which contains just the first graph is skipped
	{
		//Print the certificate steps performed to create each line
		fprintf(outFile," %s",currentLine->description);
		currentLine = currentLine->next;
	}
	fprintf(outFile," }");
}
/*
cert copyCert(cert inpCert)
{
	int i = 0;
	cert newCert;
	line currentLine = inpCert->head;
	for(i = 0; currentLine != NULL; i++)
	{
		addNewLine(newCert, copyLine(currentLine));
		currentLine = currentLine->next;
	}
	
	newCert->graph = copyGraph(inpCert->graph);
	newCert->graphCounter = inpCert->graphCounter;
	
	return newCert;
}

line copyLine(line inpLine)
{
	
	int i;
	for(i = 0; i < inpLine->numGraphs; i++)
	{
		newLine->signs[i] =  inpLine->signs[i];
		newLine->graphs[i] = 
		//Don't print the "+" for the first graph in the line
		if(i == 0 && inpLine->signs[i] == 1)
		{
			printf("G%d%s",getID(inpLine->graphs[i]),getDesc(inpLine->graphs[i]));
		}
		else
		{
			if(inpLine->signs[i] == 1)
			{
				printf(" + ");
			}
			else
			{
				printf(" - ");
			}
			printf("G%d%s",getID(inpLine->graphs[i]),getDesc(inpLine->graphs[i]));
		}
	}
	return newLine;
}*/

/*char * printCertToString(cert inpCert)
{
	char certString[500];
	char temp[500];
	int i;
	line currentLine = inpCert->head->next;
	sprintf(temp,"\nCertificate Length: %d\n", inpCert->length);
	strcat(certString,temp);
	/*fprintf(outFile,"G   = ");
	for(i = 0; i < inpCert->length; i++) //Note that the first line which contains just the first graph is skipped
	{
		//Print each line in the Certificate
		if(i!=0)	fprintf(outFile,"    =");
		fprintf(outFile,"\t");
		printLineToFile(currentLine, outFile);
		fprintf(outFile,"\n");
		currentLine = currentLine->next;
	}
	fprintf(outFile,"\n");
}*/
/******************************************************************************
Function:	printLine
Description:	Prints the given line
Inputs:		line - the line to print
Return Value:	None
Side Effects:	None
Created:	Z.Bukovac Feb 2012
******************************************************************************/
void printLine(line inpLine)
{
	int i,x;
	for(i = 0; i < inpLine->numGraphs; i++)
	{
		//Don't print the "+" for the first graph in the line
		if(i == 0 && inpLine->signs[i] == 1)
		{
			printf("G%d%s",getID(inpLine->graphs[i]),getDesc(inpLine->graphs[i]));
		}
		else
		{
			if(inpLine->signs[i] == 1)
			{
				printf(" + ");
			}
			else
			{
				printf(" - ");
			}
			printf("G%d%s",getID(inpLine->graphs[i]),getDesc(inpLine->graphs[i]));
		}
	}
}

void printLineToFile(line inpLine, FILE *outFile)
{
	int i,x;
	for(i = 0; i < inpLine->numGraphs; i++)
	{
		//Don't print the "+" for the first graph in the line
		if(i == 0 && inpLine->signs[i] == 1)
		{
			fprintf(outFile,"G%d%s",getID(inpLine->graphs[i]),getDesc(inpLine->graphs[i]));
		}
		else
		{
			if(inpLine->signs[i] == 1)
			{
				fprintf(outFile," + ");
			}
			else
			{
				fprintf(outFile," - ");
			}
			fprintf(outFile,"G%d%s",getID(inpLine->graphs[i]),getDesc(inpLine->graphs[i]));
		}
	}
	
}

void printLineCSToFile(line inpLine, FILE *outFile)
{
	int i;
	for(i = 0; i < inpLine->numGraphs; i++)
	{
		//Don't print the "+" for the first graph in the line
		if(i == 0 && inpLine->signs[i] == 1)
		{
			fprintf(outFile,"G%d%s",getID(inpLine->graphs[i]),getDesc(inpLine->graphs[i]));
		}
		else
		{
			if(inpLine->signs[i] == 1)
			{
				fprintf(outFile," + ");
			}
			else
			{
				fprintf(outFile," - ");
			}
			fprintf(outFile,"G%d%s",getID(inpLine->graphs[i]),getDesc(inpLine->graphs[i]));
		}
	}
	fprintf(outFile,"\t%s",inpLine->description);
}

/******************************************************************************
Function:	printCertGraphs
Description:	Prints all of the graphs in the last line of the given certificate.
Inputs:		cert - the certificate to print
Return Value:	None
Side Effects:	None
Created:	Z.Bukovac Feb 2012
******************************************************************************/
void printCertGraphs(cert inpCert)
{
	int i;
	line lastLine = inpCert->tail;
	
	for(i = 0; i < lastLine->numGraphs; i++)
	{
		printGraph(lastLine->graphs[i]);
	}
}

/******************************************************************************
Function:	printCertGraphs
Description:	Prints all of the graphs in the last line of the given certificate
		to the given FILE
Inputs:		cert - the certificate to print
Return Value:	None
Side Effects:	None
Created:	Z.Bukovac Feb 2012
******************************************************************************/
void printCertGraphsToFile(cert inpCert, FILE *outFile)
{
	int i;
	line lastLine = inpCert->tail;
	
	for(i = 0; i < lastLine->numGraphs; i++)
	{
		printGraphToFile(lastLine->graphs[i], outFile);
	}
}

/******************************************************************************
Function:	printOperation
Description:	Prints the details of the given operation
Inputs:		op -  the operation to print
Return Value:	None
Side Effects:	None
Created:	Z.Bukovac Feb 2012
******************************************************************************/
void printOperation(operation op)
{
	line line = op->cert->tail;
	
	// printf("\n");
	if(op->optype == DELCONT)
	{
		printf("Delete Contract G%d on edge (%d %d)", getID(line->graphs[op->graphIndex1]), op->u, op->v);
	}
	else if(op->optype ==  INVDELCONT1)
	{
		printf("Inverse Delete Contract Type 1 with G%d(Ord:%d,Sign:%d) and G%d(Ord:%d,Sign:%d) on edge (%d %d)",
			getID(line->graphs[op->graphIndex1]),getSize(line->graphs[op->graphIndex1]),op->cert->tail->signs[op->graphIndex1],getID(line->graphs[op->graphIndex2]),getSize(line->graphs[op->graphIndex2]),op->cert->tail->signs[op->graphIndex2], op->u, op->v);
	}
	else if(op->optype ==  INVDELCONT2) 
	{
		printf("Inverse Delete Contract Type 2 with G%d(Ord:%d,Sign:%d) and G%d(Ord:%d,Sign:%d) on edge (%d %d)",
			getID(line->graphs[op->graphIndex1]),getSize(line->graphs[op->graphIndex1]),op->cert->tail->signs[op->graphIndex1],getID(line->graphs[op->graphIndex2]),getSize(line->graphs[op->graphIndex2]),op->cert->tail->signs[op->graphIndex2], op->u, op->v);
	}
	else if(op->optype ==  ADDIDENT)
	{
		printf("Add Identify G%d on edge (%d %d)",getID(line->graphs[op->graphIndex1]), op->u, op->v);
	}
	else if(op->optype ==  INVADDIDENT) 
	{
		printf("Inverse Add Identify G%d(Ord:%d,Sign:%d) and G%d(Ord:%d,Sign:%d) on edge (%d %d)",
			getID(line->graphs[op->graphIndex1]),getSize(line->graphs[op->graphIndex1]),op->cert->tail->signs[op->graphIndex1],getID(line->graphs[op->graphIndex2]),getSize(line->graphs[op->graphIndex2]),op->cert->tail->signs[op->graphIndex2], op->u, op->v);
	}
	else if(op->optype ==  CANCEL) 
	{
		printf("Cancel G%d and G%d",
			getID(line->graphs[op->graphIndex1]),getID(line->graphs[op->graphIndex2]));
	}
}

/******************************************************************************
Function:	addNewLine
Description:	Adds the given line to the tail of the given certificate
Inputs:		cert - the certificate to add to
		line - the line to be addedd to the current tail of the cert
Return Value:	None
Side Effects:	Alters cert by adding a new line
Created:	Z.Bukovac Feb 2012
******************************************************************************/
void addNewLine(cert inpCert, line inpLine)
{
	if(inpCert->head == NULL)
	{
		inpCert->head = inpLine;
		inpCert->tail = inpLine;
		inpLine->next = NULL;
		inpLine->prev = NULL;
	}
	else
	{
		inpLine->next = NULL;
		inpLine->prev = inpCert->tail;
		inpCert->tail->next = inpLine;
		inpCert->tail = inpLine;
	}
	inpCert->length++;
}


/******************************************************************************
Function:	removeLastLine
Description:	Removes the current tail of the certificate
Inputs:		cert - the certificate to remove the final line from
Return Value:	None
Side Effects:	Alters cert by removing the tail line
Created:	Z.Bukovac Feb 2012
******************************************************************************/
void removeLastLine(cert inpCert)
{
	int graphsFreed = 0;
	
	if(inpCert->tail == NULL)
	{
		graphsFreed = freeLine(inpCert->tail);
		inpCert->tail = NULL;	
	}
	else
	{
		inpCert->tail = inpCert->tail->prev;
		graphsFreed = freeLine(inpCert->tail->next);
		inpCert->tail->next = NULL;
	}
	
	inpCert->graphCounter -= graphsFreed;
	
	//Reduce certificate length
	inpCert->length--;
}

int freeLine(line inpLine)
{
	int i = 0, graphsFreed = 0;
	
	for(i = 0; i < inpLine->numGraphs; i++)
	{	
		//Destroy this reference to the graph;
		destroyGraphRef(&inpLine->graphs[i]);
		
		if(inpLine->graphs[i] == NULL)
		{
			graphsFreed++;
		}
	}
	free(inpLine->graphs);
	free(inpLine->signs);
	free(inpLine);
	return graphsFreed;
}

/*
void freeLineUnconditional(line inpLine)
{
	int i = 0, graphsFreed = 0;
	graph currentGraph;
	for(i = 0; i < inpLine->numGraphs; i++)
	{
		freeGraph(inpLine->graphs[i]);
	}
	free(inpLine);
	return graphsFreed;
}*/

void freeCert(cert inpCert)
{
	int i = 0;
	destroyGraphRef(&inpCert->graph);
	line currentLine = inpCert->head;
	line nextLine = NULL;
	while(currentLine != NULL)
	{
		nextLine = currentLine->next;
		freeLine(currentLine);
		currentLine = nextLine;
	}
	free(inpCert);
}

void freeList(operationList inpList)
{
	int i = 0;
	operationListNode currentNode = inpList->head;
	operationListNode nextNode;
	
	while(currentNode != NULL)
	{
		free(currentNode->op);
		
		nextNode = currentNode->next;
		free(currentNode);
		currentNode = nextNode;
	}
	free(inpList);
}

line getCurrentLine(cert inpCert)
{
	return inpCert->tail;
}



/******************************************************************************
Function:	performOp
Description:	Calls the appropriate function to perform the requested operation on the certificate
Inputs:		op - the operation to be performed
Return Value:	None
Side Effects:	Alters cert by performing the operation requested
Created:	Z.Bukovac Aug 2012
******************************************************************************/
void performOp(operation op)
{
	//bool valid = true;
	//printOperation(op);
	switch(op->optype)
	{
		case DELCONT:		deleteCont(op->cert, op->graphIndex1, op->u, op->v);				break;
		case ADDIDENT:		addIdent(op->cert, op->graphIndex1, op->u, op->v);				break;
		case INVDELCONT1:	deleteContRev1(op->cert, op->graphIndex1, op->graphIndex2, op->u, op->v);	break;
		case INVDELCONT2:	deleteContRev2(op->cert, op->graphIndex1, op->graphIndex2, op->u, op->v);	break;
		case INVADDIDENT:	addIdentRev(op->cert, op->graphIndex1, op->graphIndex2, op->u, op->v);		break;
		case CANCEL:		cancelGraphs(op->cert, op->graphIndex1, op->graphIndex2);			break;
		default:												break;
	}
}

/******************************************************************************
Function:	deleteCont
Description:	Adds a new line to the cert which results from performing a deletion-contraction
		on the graph at graphIndex in the tail line of the given certificate.
Inputs:		cert - the certificate containing the relevant graph for the deletion-contraction
		graphIndex - indicates which graph in the tail line of the given certificate
		u,v - indicates the edge to delete-contract on
Return Value:	None
Side Effects:	Alters cert by adding a new line to the tail.
Created:	Z.Bukovac Feb 2012
******************************************************************************/
void deleteCont(cert inpCert, int graphIndex, int u, int v)
{	
	int i = 0, j = 0;
	
	//printf("\nAttempting Deletion Contraction on G%d(%d,%d)...",getID(inpCert->tail->graphs[graphIndex]),u,v);
	line nextLine = newLine();
	line oldLine = inpCert->tail;
	nextLine->numGraphs = oldLine->numGraphs+1;
	
	nextLine->graphs = malloc(nextLine->numGraphs * sizeof(graph));
	nextLine->signs = malloc(nextLine->numGraphs * sizeof(int));
	
	for(i = 0; i < oldLine->numGraphs; i++)
	{
		//If this is the graph we need to operate on...
		if(i == graphIndex)
		{
			//Del Cont
			nextLine->graphs[j] =  copyAndDeleteEdge(oldLine->graphs[i],u,v);
			nextLine->graphs[j+1] = copyAndContractEdge(oldLine->graphs[i],u,v);

			//printf("\n!!!!!!!!!!!!!!! %d !!!!!!!!!!!!!!!!!!\n", inpCert->graphCounter);
			setID(nextLine->graphs[j], inpCert->graphCounter);
			setID(nextLine->graphs[j+1], inpCert->graphCounter+1);

			//Adjust the signs of the two new graphs
			nextLine->signs[j] = oldLine->signs[i];
			nextLine->signs[j+1] = oldLine->signs[i]*-1;

			//Move the index for the newLine past the two new graphs
			j += 2;
		}
		else
		{
			//Otherwise, just point the newLine's graph array at the graphs that aren't changing
			nextLine->graphs[j] = createGraphRef(oldLine->graphs[i]);
			nextLine->signs[j] = oldLine->signs[i];
			j++;
		}
	}

	sprintf(nextLine->description,"(CS1)");
	inpCert->graphCounter+=2;
	addNewLine(inpCert, nextLine);
	//printf("New line added to certificate.\n");
}


/******************************************************************************
Function:	addIdent
Description:	Adds a new line to the cert which results from performing an addition-identification
		on the graph at graphIndex in the tail line of the given certificate.
Inputs:		cert - the certificate containing the relevant graph for the addition-identification
		graphIndex - indicates which graph in the tail line of the given certificate
		u,v - indicates the edge to add-identify on
Return Value:	None
Side Effects:	Alters cert by adding a new line to the tail.
Created:	Z.Bukovac Feb 2012
******************************************************************************/
void addIdent(cert inpCert, int graphIndex, int u, int v)
{
	int i = 0, j = 0;
	
	//printf("\nAttempting Addition Identification on G%d(%d,%d)",getID(inpCert->tail->graphs[graphIndex]),u,v);
	line nextLine = newLine();
	line oldLine = inpCert->tail;
	nextLine->numGraphs = oldLine->numGraphs+1;
	
	nextLine->graphs = malloc(nextLine->numGraphs * sizeof(graph));
	nextLine->signs = malloc(nextLine->numGraphs * sizeof(int));
	
	for(i = 0; i < oldLine->numGraphs; i++)
	{
		//If this is the graph we need to operate on...
		if(i == graphIndex)
		{
			//
			nextLine->graphs[j] =copyAndAddEdge(oldLine->graphs[i],u,v);
			nextLine->graphs[j+1] = copyAndContractEdge(oldLine->graphs[i],u,v);

			setID(nextLine->graphs[j], inpCert->graphCounter);
			setID(nextLine->graphs[j+1], inpCert->graphCounter+1);

			//Adjust the signs of the two new graphs
			nextLine->signs[j] = oldLine->signs[i];
			nextLine->signs[j+1] = oldLine->signs[i];


			//Move the index for the newLine past the two new graphs
			j += 2;
		}
		else
		{
			//Otherwise, just point the newLine's graph array at the graphs that aren't changing
			nextLine->graphs[j] = createGraphRef(oldLine->graphs[i]);
			nextLine->signs[j] = oldLine->signs[i];
			j++;
		}
	}

	sprintf(nextLine->description,"(CS3)");
	inpCert->graphCounter+=2;
	addNewLine(inpCert, nextLine);
	//printf("New line added to certificate.\n");
}

/******************************************************************************
Function:	delContRev1
Description:	Adds a new line to the cert which results from performing an deletion-contraction reversal type 1
		on the graphs in the tail line of the given certificate.
Inputs:		cert - the certificate containing the relevant graph for the reversal
		graphIndex1, graphIndex2 - the graphs involved in the op
		u,v - indicates the edge involved in the reversal
Return Value:	None
Side Effects:	Alters cert by adding a new line to the tail.
Notes:		This function currently expects that the graph at cert->tail->graphs[graphIndex1] and
		cert->tail->graphs[graphIndex2] have a valid reversal. Bad things may happen otherwise.
Created:	Z.Bukovac Jun 2012
******************************************************************************/
void deleteContRev1(cert inpCert, int graphIndex1, int graphIndex2, int u, int v)
{
	int i = 0, j = 0;		//i is the oldLine's index, j is the newLine's index
	graph larger, smaller;
	int largerIndex;
	graph g1 = inpCert->tail->graphs[graphIndex1];
	graph g2 = inpCert->tail->graphs[graphIndex2];
	
	//printf("\nAttempting Deletion Contraction Reversal on G%d and G%d at (%d,%d)...",
		//getID(inpCert->tail->graphs[graphIndex1]),getID(inpCert->tail->graphs[graphIndex2]),u,v);
		
	//Set up the new line in the certificate
	line nextLine = newLine();
	line oldLine = inpCert->tail;
	nextLine->numGraphs = oldLine->numGraphs-1;
	
	nextLine->graphs = malloc(nextLine->numGraphs * sizeof(graph));
	nextLine->signs = malloc(nextLine->numGraphs * sizeof(int));
	

	//Determine which is the larger graph
	if(getSize(g1) == getSize(g2)+1)
	{
		larger = g1;
		smaller = g2;
		largerIndex = graphIndex1;
	}
	else if(getSize(g1)+1 == getSize(g2))
	{
		larger = g2;
		smaller = g1;
		largerIndex = graphIndex2;
	}
	
	for(i = 0; i < oldLine->numGraphs; i++)
	{
		//printf("\n0 i:%d, oldGcount:%d, gind1:%d, gind2:%d",i, oldLine->numGraphs,graphIndex1 ,graphIndex2);
		
		if(i == graphIndex2 && i+1 < oldLine->numGraphs)
		{
			//printf("\ninc i");
			//This graph will be removed in the nextLine, so increment the oldLine's current index
			i++;
		}
		//If this is the first graph to operate on
		if(i == graphIndex1)
		{
			//printf("First Index...creating graph G%d...",inpCert->graphCounter);
			//The new graph goes here
			nextLine->graphs[j] =  copyAndAddEdge(larger,u,v);
			//TODO add int the details for the other two reversals here, and they're done too...Don't forget signs though.

			setID(nextLine->graphs[j], inpCert->graphCounter);

			//The new graph gets the same sign as the larger graph;
			nextLine->signs[j] = oldLine->signs[largerIndex];

			//Move the index for the newLine past the new graph
			j++;
		}
		if(i != graphIndex1 && i != graphIndex2)
		{
			//printf("\ncopy and inc j");
			//Otherwise, just point the nextLine's graph array at the graphs that aren't changing
			nextLine->graphs[j] = createGraphRef(oldLine->graphs[i]);
			nextLine->signs[j] = oldLine->signs[i];
			j++;
		}
		//printf("\n1 i:%d, oldGcount:%d, gind1:%d, gind2:%d",i, oldLine->numGraphs,graphIndex1 ,graphIndex2);
	}

	sprintf(nextLine->description,"(CS2)");
	inpCert->graphCounter+=1;
	addNewLine(inpCert, nextLine);
	//printf("New line added to certificate.\n");
}

void deleteContRev2(cert inpCert, int graphIndex1, int graphIndex2, int u, int v)
{
	int i = 0, j = 0;		//i is the oldLine's index, j is the newLine's index
	graph larger, smaller;
	int smallerIndex;
	graph g1 = inpCert->tail->graphs[graphIndex1];
	graph g2 = inpCert->tail->graphs[graphIndex2];
	
	//printf("\nAttempting Deletion Contraction Reversal on G%d and G%d at (%d,%d)...",
		//getID(inpCert->tail->graphs[graphIndex1]),getID(inpCert->tail->graphs[graphIndex2]),u,v);
		
	//Set up the new line in the certificate
	line nextLine = newLine();
	line oldLine = inpCert->tail;
	nextLine->numGraphs = oldLine->numGraphs-1;
	
	nextLine->graphs = malloc(nextLine->numGraphs * sizeof(graph));
	nextLine->signs = malloc(nextLine->numGraphs * sizeof(int));
	
	
	//Determine which is the graph with more edges
	if(getNumEdges(g1) == getNumEdges(g2)+1)
	{
		larger = g1;
		smaller = g2;
		smallerIndex = graphIndex2;
	}
	else if(getNumEdges(g1)+1 == getNumEdges(g2))
	{
		larger = g2;
		smaller = g1;
		smallerIndex = graphIndex1;
	}
	
	for(i = 0; i < oldLine->numGraphs; i++)
	{
		//printf("\n0 i:%d, oldGcount:%d, gind1:%d, gind2:%d",i, oldLine->numGraphs,graphIndex1 ,graphIndex2);
		
		if(i == graphIndex2 && i+1 < oldLine->numGraphs)
		{
			//printf("\ninc i");
			//This graph will be removed in the nextLine, so increment the oldLine's current index
			i++;
		}
		//If this is the first graph to operate on
		if(i == graphIndex1)
		{
			//printf("First Index...creating graph G%d...",inpCert->graphCounter);
			//The new graph goes here
			nextLine->graphs[j] =  copyAndContractEdge(smaller,u,v);
			//TODO add int the details for the other two reversals here, and they're done too...Don't forget signs though.

			setID(nextLine->graphs[j], inpCert->graphCounter);

			//The new graph gets the same sign as the larger graph;
			nextLine->signs[j] = oldLine->signs[smallerIndex];

			//Move the index for the newLine past the new graph
			j++;
		}
		if(i != graphIndex1 && i != graphIndex2)
		{
			//printf("\ncopy and inc j");
			//Otherwise, just point the nextLine's graph array at the graphs that aren't changing
			nextLine->graphs[j] = createGraphRef(oldLine->graphs[i]);
			nextLine->signs[j] = oldLine->signs[i];
			j++;
		}
		//printf("\n1 i:%d, oldGcount:%d, gind1:%d, gind2:%d",i, oldLine->numGraphs,graphIndex1 ,graphIndex2);
	}

	sprintf(nextLine->description,"(CS5)");
	inpCert->graphCounter+=1;
	addNewLine(inpCert, nextLine);
	//printf("New line added to certificate.\n");
}

void addIdentRev(cert inpCert, int graphIndex1, int graphIndex2, int u, int v)
{
	int i = 0, j = 0;		//i is the oldLine's index, j is the newLine's index
	graph larger, smaller;
	int largerIndex;
	graph g1 = inpCert->tail->graphs[graphIndex1];
	graph g2 = inpCert->tail->graphs[graphIndex2];
	
	//printf("\nAttempting Deletion Contraction Reversal on G%d and G%d at (%d,%d)...",
		//getID(inpCert->tail->graphs[graphIndex1]),getID(inpCert->tail->graphs[graphIndex2]),u,v);
		
	//Set up the new line in the certificate
	line nextLine = newLine();
	line oldLine = inpCert->tail;
	nextLine->numGraphs = oldLine->numGraphs-1;
	
	nextLine->graphs = malloc(nextLine->numGraphs * sizeof(graph));
	nextLine->signs = malloc(nextLine->numGraphs * sizeof(int));
	

	//Determine which is the larger graph
	if(getSize(g1) == getSize(g2)+1)
	{
		larger = g1;
		smaller = g2;
		largerIndex = graphIndex1;
	}
	else if(getSize(g1)+1 == getSize(g2))
	{
		larger = g2;
		smaller = g1;
		largerIndex = graphIndex2;
	}
	
	for(i = 0; i < oldLine->numGraphs; i++)
	{
		//printf("\n0 i:%d, oldGcount:%d, gind1:%d, gind2:%d",i, oldLine->numGraphs,graphIndex1 ,graphIndex2);
		
		if(i == graphIndex2 && i+1 < oldLine->numGraphs)
		{
			//printf("\ninc i");
			//This graph will be removed in the nextLine, so increment the oldLine's current index
			i++;
		}
		//If this is the first graph to operate on
		if(i == graphIndex1)
		{
			//printf("First Index...creating graph G%d...",inpCert->graphCounter);
			//The new graph goes here
			nextLine->graphs[j] =  copyAndDeleteEdge(larger,u,v);
			//TODO add int the details for the other two reversals here, and they're done too...Don't forget signs though.

			setID(nextLine->graphs[j], inpCert->graphCounter);

			//The new graph gets the same sign as the larger graph;
			nextLine->signs[j] = oldLine->signs[largerIndex];

			//Move the index for the newLine past the new graph
			j++;
		}
		if(i != graphIndex1 && i != graphIndex2)
		{
			//printf("\ncopy and inc j");
			//Otherwise, just point the nextLine's graph array at the graphs that aren't changing
			nextLine->graphs[j] = createGraphRef(oldLine->graphs[i]);
			nextLine->signs[j] = oldLine->signs[i];
			j++;
		}
		//printf("\n1 i:%d, oldGcount:%d, gind1:%d, gind2:%d",i, oldLine->numGraphs,graphIndex1 ,graphIndex2);
	}

	sprintf(nextLine->description,"(CS4)");
	inpCert->graphCounter+=1;
	addNewLine(inpCert, nextLine);
	//printf("New line added to certificate.\n");
}


/******************************************************************************
Function:	willCancel
Description:	Checks if the two graphs will cancel
Inputs:		line - the line containing the two graphs
		graphIndex1, graphIndex2 - the graphs involved
Return Value:	Returns 1 if they cancel, 0 if not.
Side Effects:	None
Notes:		
Created:	Z.Bukovac Feb 2012
******************************************************************************/
bool willCancel(line inpLine, int graphIndex1, int graphIndex2)
{
	graph g1 = inpLine->graphs[graphIndex1];
	graph g2 = inpLine->graphs[graphIndex2];
	
	//If they are of the same order, have opposing signs and are also isomorpic...
	if(getSize(g1) == getSize(g2) && 
	   inpLine->signs[graphIndex1] != inpLine->signs[graphIndex2] && 
	   isomorphic(g1,g2))
	{
		//Then they can cancel
		return true;
	}
	
	//They're not cancelable
	return false;
}


/******************************************************************************
Function:	getCancelOperationList()
Description:	Returns an operationList of containing all possible cancelling
		operations in the tail line of the inpCert.
Inputs:		inpCert - the certificate in question
		graphIndex1, graphIndex2 - the indicies of graphs involved
Return Value:	list - a new operationList containing all cancel ops in the tail
Side Effects:	None
Notes:		
Created:	Z.Bukovac Aug 2012
******************************************************************************/
operationList getCancelOperationList(cert inpCert)
{	
	operationList list = newOperationList();

	int i = 0,j = 0;

	//Check all pairs of graphs in the the tail of the certificate
	for(i = 0; i < inpCert->tail->numGraphs; i++)
	{
		for(j = 0; j < i; j++)
		{
			if(willCancel(inpCert->tail,i,j))
			{
				//If this pair will cancel, add a new operation to the list
				addOpToList(list, newOperation(inpCert, CANCEL, i, j, 0, 0));
			}
		}
	}
	
	return list;
}

/******************************************************************************
Function:	printCancellableGraphPairs
Description:	Prints all of the graph pairs that will cancel in the given ‌line
Inputs:		line - the line containing the graphs
Return Value:	None
Side Effects:	None
Notes:		TODO: Probably Going...
Created:	Z.Bukovac Feb 2012
******************************************************************************/
void printCancelableGraphPairs(cert inpCert)
{
	int i = 0, j = 0, x = 0;
	printf("\nCancelable Graph Pairs: ");
	for(i = 0; i < inpCert->tail->numGraphs; i++)
	{
		for(j = i+1; j < inpCert->tail->numGraphs; j++)
		{
			if(willCancel(inpCert->tail,i,j))
			{
				//Print the pairs that cancel
				printf("\n1: G%d and G%d",getID(inpCert->tail->graphs[i]),getID(inpCert->tail->graphs[j]));
				x = 1;
			}
		}
	}
	if(x == 0)
	{
		printf("There are no cancelable pairs");
	}
}

/******************************************************************************
Function:	cancelGraphs
Description:	Adds a new line to the cert which results from performing a cancelation on
		the graphs in the tail
Inputs:		line - the line containing the graphs
Return Value:	None
Side Effects:	None
Notes:		
Created:	Z.Bukovac Feb 2012
******************************************************************************/
void cancelGraphs(cert inpCert, int graphIndex1, int graphIndex2)
{
	//printf("\nAttempting to cancel G%d and G%d",getID(inpCert->tail->graphs[graphIndex1]),getID(inpCert->tail->graphs[graphIndex2]));
	line nextLine = newLine();
	line oldLine = inpCert->tail;
	nextLine->numGraphs = oldLine->numGraphs-2;		//Account for the two graphs that are being canceled
	
	nextLine->graphs = malloc(nextLine->numGraphs * sizeof(graph));
	nextLine->signs = malloc(nextLine->numGraphs * sizeof(int));
	
	int i = 0, j = 0;
	
	//Copy everything except the two graphs to be cancelled
	for(i = 0; i < oldLine->numGraphs; i++)
	{	
		//If the current isn't that of one of the graphs being removed...
		if(i != graphIndex1 && i != graphIndex2)
		{
			//Point the newLine's graph array at the graphs that aren't changing
			nextLine->graphs[j] = createGraphRef(oldLine->graphs[i]);
			nextLine->signs[j] = oldLine->signs[i];
			j++;
		}
	}

	sprintf(nextLine->description,"(CS8)");
	addNewLine(inpCert, nextLine);
}

/******************************************************************************
Function:	willDC2Reverse
Description:	Determines if the given graphs will DC2 reverse
Inputs:		line - the line containing the graphs
		graphIndex1, graphIndex2: the indices of the graphs in question
Return Value:	int  TODO: Must return an operation struct eventually.
Side Effects:	None
Notes:		TODO Not yet working...
		
Created:	Z.Bukovac Apr 2012
******************************************************************************/
int willDC2Reverse(line inpLine, int graphIndex1, int graphIndex2)
{
	//The smaller graph needs to be negative
	graph g1 = inpLine->graphs[graphIndex1];
	graph g2 = inpLine->graphs[graphIndex2];
	graph temp;
	
	graph larger, smaller;
	int i,j = 0;

	//Determine if their orders are compatible, and which one, if either, is larger.
	if(getSize(g1) == getSize(g2)+1)
	{
		larger = g1;
		smaller = g2;
	}
	else if(getSize(g1)+1 == getSize(g2))
	{
		larger = g2;
		smaller = g1;
	}
	else
	{
		//Their orders are incompatible for a reverse DC1 op.
		return 0;
	}
	
	//If they have opposing signs...
	if(inpLine->signs[graphIndex1] != inpLine->signs[graphIndex2])
	{
		//Search through the non-edges that when contracted in the larger graph produce the smaller graph.
		//For each non-edge in the graph...
			//Try contracting on its vertices and then check if (contract(larger)==smaller)
			//If so, it's reversable, so return a new operation stuct with the relevant details.

		for(i = 0; i < getSize(larger); i++)
		{
			for(j = 0; j < i; j++)
			{
				//If there is no edge at ij...
				if(!isEdge(larger,i,j))
				{
					//Contract on i and j, then compare to the smaller graph
					temp = copyAndContractEdge(larger,i,j);
					if(isomorphic(temp,smaller))
					{
						//TODO They are reversable, so return a new operation stuct with the relevant details.
						printf("\nG%d and G%d reversable on on edge (%d,%d) of G%d",getID(larger),getID(smaller),i,j,getID(larger));

					}
					freeGraph(temp);
				}
			}
		}
		return 1;
	}
	
	//They're not reverseable, so return null;
	return 0;
}

int willAIReverse(line inpLine, int graphIndex1, int graphIndex2)
{
	graph g1 = inpLine->graphs[graphIndex1];
	graph g2 = inpLine->graphs[graphIndex2];
	//If they are of the same order, have opposing signs and are also isomorpic...
	if(getSize(g1) == getSize(g2) || getSize(g1) == getSize(g2)+1 || getSize(g1)+1 == getSize(g2) && 
	   inpLine->signs[graphIndex1] != inpLine->signs[graphIndex2] && 
	   isomorphic(g1,g2))
	{
		//Then they can cancel
		return 1;
	}
	
	//They're not cancelable
	return 0;
}

void printCurrentGraphList(cert inpCert)
{
	int i;
	
	for(i = 0; i < inpCert->tail->numGraphs; i++)
	{
		printf("\t%d: G%d\n", i, getID(inpCert->tail->graphs[i]));
	}
}

int getCurrentLineLength(cert inpCert)
{
	return inpCert->tail->numGraphs;
}

int getLength(cert inpCert)
{
	return inpCert->length;
}

graph getGraph(cert inpCert, int index)
{
	return inpCert->tail->graphs[index];
}

int getIdAtIndex(cert inpCert, int index)
{
	return getID(inpCert->tail->graphs[index]);
}

bool canDeleteCont(cert inpCert, int graphIndex, int u, int v)
{
	return isEdge(inpCert->tail->graphs[graphIndex], u, v);
}

bool canAddIdent(cert inpCert, int graphIndex, int u, int v)
{
	return !isEdge(inpCert->tail->graphs[graphIndex], u, v) 	&& 
	       isValidVertex(inpCert->tail->graphs[graphIndex], u)	&& 
	       isValidVertex(inpCert->tail->graphs[graphIndex], v)	;
}



/******************************************************************************
Function:	getDC1RevOperationList
Description:	Returns a list of new operations for any pairs of graphs that
		can be found in the inpCert->tail that will DC1 reverse.
Inputs:		inpCert - the certificate to look at (specifically, its tail line)
Return Value:	list - contains all DC1 reversable graph pair ops in the cert->tail
		If there are no such reversal operations, the list is still
		returned, but it is empty.
Side Effects:	None
Notes:		
Created:	Z.Bukovac Feb 2012
******************************************************************************/
operationList getDC1RevOperationList(cert inpCert)
{	
	operationList list = newOperationList();

	int g1 = 0, g2 = 0;
	int i = 0,j = 0;
	bool ordersOk = false;
	operation op;
	graph larger, smaller, temp;
	int largerIndex, smallerIndex;		//The index of the larger graph

	//Check all pairs of graphs and get the possible reversals for each
	for(g1 = 0; g1 < inpCert->tail->numGraphs; g1++)
	{
		for(g2 = 0; g2 < g1; g2++)
		{	
			//The smaller graph needs to be negative
			//graph g1 = inpCert->tail->graphs[g1];
			//graph g2 = inpCert->tail->graphs[g2];
			//printf("\nINVDELCONT1 Test G%d & G%d...",getID(inpCert->tail->graphs[g1]),getID(inpCert->tail->graphs[g2]));


			//Determine if their orders are compatible, and which one, if either, is larger.
			if(getSize(inpCert->tail->graphs[g1]) == getSize(inpCert->tail->graphs[g2])+1)
			{
				larger = inpCert->tail->graphs[g1];
				smaller = inpCert->tail->graphs[g2];
				ordersOk = true;
				//printf("orders ok...");
			}
			else if(getSize(inpCert->tail->graphs[g1])+1 == getSize(inpCert->tail->graphs[g2]))
			{
				larger = inpCert->tail->graphs[g2];
				smaller = inpCert->tail->graphs[g1];
				ordersOk = true;
				//printf("orders ok...");
			}
			else
			{
				//Their orders are incompatible for a reverse DC1 op.
				//printf("orders off...");
				ordersOk = false;
			}
	
			//If they have opposing signs...
			//printf("%d and %d :%d", inpCert->tail->signs[g1],inpCert->tail->signs[g2], ordersOk && inpCert->tail->signs[g1] != inpCert->tail->signs[g2]);
			if(ordersOk && inpCert->tail->signs[g1] != inpCert->tail->signs[g2])
			{
				//printf("signs ok...");
				//Search through the non-edges that when contracted in the larger graph produce the smaller graph.
				//For each non-edge in the graph...
					//Try contracting on its vertices and then check if (contract(larger) == smaller)
					//If so, it's reversable, so return a new operation stuct with the relevant details.

				for(i = 0; i < getSize(larger); i++)
				{
					for(j = 0; j < i; j++)
					{
						//If there is no edge at ij...
						if(!isEdge(larger,i,j))
						{
							//printf("; edge (%d,%d)...",i,j);
							
							//DIAG printGraph(copyAndContractEdge(larger,i,j));
							/*DIAG: It seems that the first index is off by one, so the new graph trying
								to add an edge at an index of graph->size (one too many) and some point
								during the following call:
							*/	
							//Contract on i and j, then compare to the smaller graph
							temp = copyAndContractEdge(larger,i,j);
							//printf("@%d,%d:",i,j); printGraph(temp); printGraphDiagnostics(temp);
							if(isomorphic(temp,smaller))
							{
							//	printf("\nARE ISO:\n=======\n");
							//	printGraph(copyAndContractEdge(larger,i,j));
							//	printGraph(smaller);
								//printf("contracts for equality, adding to list!");
								//TODO They are reversable, so add a new operation 
								addOpToList(list, newOperation(inpCert, INVDELCONT1, g1, g2, i, j));
								//printf("%d,%d GRAPHS: %d & %d INVDELCONT1\*\n",i,j,getID(inpCert->tail->graphs[g1]),getID(inpCert->tail->graphs[g2]));
							}
							freeGraph(temp);
						}
					}
				}
			}
		}
	}
	
	return list;
}

operationList getDC2RevOperationList(cert inpCert)
{	
	operationList list = newOperationList();

	int graphIndex1 = 0, graphIndex2 = 0;
	int i = 0,j = 0;
	bool ordersOk = false;
	operation op;
	graph larger, smaller, temp;
	int largerIndex, smallerIndex;		

	//Check all pairs of graphs and get the possible reversals for each
	for(graphIndex1 = 0; graphIndex1 < inpCert->tail->numGraphs; graphIndex1++)
	{
		for(graphIndex2 = 0; graphIndex2 < graphIndex1; graphIndex2++)
		{	
			//The smaller graph needs to be negative
			graph g1 = inpCert->tail->graphs[graphIndex1];
			graph g2 = inpCert->tail->graphs[graphIndex2];
			//printf("\nINVDELCONT2 Test G%d & G%d...",getID(g1),getID(g2));
			

			//Determine if sizes and edge counts are compatible, and which one, if either, has more edges (aka is "larger").
			if(getSize(g1) == getSize(g2) && getNumEdges(g1) == getNumEdges(g2)+1)
			{
				larger = g1;
				smaller = g2;
				largerIndex = graphIndex1;
				smallerIndex = graphIndex2;
				//printf("orders ok...");
			}
			else if(getSize(g1) == getSize(g2) && getNumEdges(g1)+1 == getNumEdges(g2))
			{
				larger = g2;
				smaller = g1;
				largerIndex = graphIndex2;
				smallerIndex = graphIndex1;
				//printf("orders ok...");
			}
			else
			{
				//Their orders are incompatible for a reverse DC1 op.
				//printf("orders off...");
				break; //Return an empty list
			}
	
			//If they have opposing signs...
			if(inpCert->tail->signs[graphIndex1] != inpCert->tail->signs[graphIndex2])
			{
				//printf("signs ok...");
				//Search through the non-edges that when contracted in the larger graph produce the smaller graph.
				//For each non-edge in the graph...
					//Try contracting on its vertices and then check if (contract(larger) == smaller)
					//If so, it's reversable, so return a new operation stuct with the relevant details.

				for(i = 0; i < getSize(larger); i++)
				{
					for(j = 0; j < i; j++)
					{
						//If there is no edge ij in the smaller...
						if(!isEdge(smaller,i,j))
						{
							//printf("; edge (%d,%d)...",i,j);
							//Add the edge to the smaller and compare to the larger
							temp = copyAndAddEdge(smaller,i,j);
							if(isomorphic(larger,temp))
							{
								//printf("\nARE ISO:\n=======\n");
							//	printGraph(copyAndContractEdge(larger,i,j));
							//	printGraph(smaller);
								//printf("edge adds for equality, adding to list!");
								//TODO They are reversable, so add a new operation 
								addOpToList(list, newOperation(inpCert, INVDELCONT2, largerIndex, smallerIndex, i, j));
							}
							freeGraph(temp);
						}
					}
				}
			}
		}
	}
	
	return list;
}


operationList getAIRevOperationList(cert inpCert)
{	
	operationList list = newOperationList();
	int g1 = 0, g2 = 0;
	int i = 0,j = 0;
	bool ordersOk = false;
	operation op;
	graph larger, smaller, temp;
	int largerIndex, smallerIndex;		//The index of the larger graph

	//Check all pairs of graphs and get the possible reversals for each
	for(g1 = 0; g1 < inpCert->tail->numGraphs; g1++)
	{
		for(g2 = 0; g2 < g1; g2++)
		{	
			//Determine if their orders are compatible, and which one, if either, is larger.
			if(getSize(inpCert->tail->graphs[g1]) == getSize(inpCert->tail->graphs[g2])+1)
			{
				larger = inpCert->tail->graphs[g1];
				smaller = inpCert->tail->graphs[g2];
				ordersOk = true;
				//printf("orders ok...");
			}
			else if(getSize(inpCert->tail->graphs[g1])+1 == getSize(inpCert->tail->graphs[g2]))
			{
				larger = inpCert->tail->graphs[g2];
				smaller = inpCert->tail->graphs[g1];
				ordersOk = true;
				//printf("orders ok...");
			}
			else
			{
				//Their orders are incompatible for a reverse DC1 op.
				//printf("orders off...");
				ordersOk = false;
			}
	
			//If their orders are ok and they have the same sign...
			if(ordersOk && inpCert->tail->signs[g1] == inpCert->tail->signs[g2])
			{
			
				//printf("signs ok...");
				//Search through the edges that when contracted in the larger graph produce the smaller graph.
				//For each edge in the graph...
					//Try contracting on its vertices and then check if (contract(larger) == smaller)
					//If so, it's reversable, so return a new operation stuct with the relevant details.

				for(i = 0; i < getSize(larger); i++)
				{
					for(j = 0; j < i; j++)
					{
						//If there is an edge at ij...
						if(isEdge(larger,i,j))
						{
							/*printf("\nFOUND:\n=======\n");
							printGraph(larger); 
							printf("\nCONTRACTS ON (%d %d)TO:\n=======\n", i, j);
							printGraph(copyAndContractEdge(larger,i,j));
							*///printf("; edge (%d,%d)...",i,j);
							//Contract on i and j, then compare to the smaller graph
							temp = copyAndContractEdge(larger,i,j);
							if(isomorphic(temp,smaller))
							{
								
								//printf("\nWHICH IS ISO TO:\n=======\n");
								//printGraph(smaller);
								//printf("contracts for equality, adding to list!");
								//They are reversable, so add a new operation 
								addOpToList(list, newOperation(inpCert, INVADDIDENT, g1, g2, i, j));
								//printf("%d,%d GRAPHS: %d & %d INVADDIDENT\*\n",i,j,getID(inpCert->tail->graphs[g1]),getID(inpCert->tail->graphs[g2]));
							}
							freeGraph(temp);
						}
					}
				}
			}
		}
	}
	
	return list;
}




operationList getPairwiseOperationList(cert inpCert)
{
	operationList list = newOperationList();
	int g1 = 0, g2 = 0;
	int i = 0,j = 0;
	operation op = NULL;
	graph larger = NULL, smaller = NULL, temp = NULL;
	//int largerIndex, smallerIndex;		//The index of the larger graph
	
	bool tryAI = false;
	bool tryDC1 = false;
	bool tryDC2 = false;


	//Check all pairs of graphs and get the possible reversals for each
	for(g1 = 0; g1 < inpCert->tail->numGraphs; g1++)
	{
		for(g2 = 0; g2 < g1; g2++)
		{	
			//printf("%d,%d GRAPHS: %d & %d\n",g1,g2,getID(inpCert->tail->graphs[g1]),getID(inpCert->tail->graphs[g2]));
			
			if(willCancel(inpCert->tail,g1,g2))
			{
				//If this pair will cancel, add a new operation to the list
				addOpToList(list, newOperation(inpCert, CANCEL, g1, g2, 0, 0));
			}
			else if(getSize(inpCert->tail->graphs[g1]) == getSize(inpCert->tail->graphs[g2])+1)
			{
				//This pair are potentially AI or DC1 invertable
				//Determine if their orders are compatible for an inverse op, and which one, if either, is larger.
				larger = inpCert->tail->graphs[g1];
				smaller = inpCert->tail->graphs[g2];
				
				if(inpCert->tail->signs[g1] == inpCert->tail->signs[g2])
				{
					tryAI = true;
				}
				//else if(inpCert->tail->signs[g1] != inpCert->tail->signs[g2])
				else{
					tryDC1 = true;
				}
				
			}
			else if(getSize(inpCert->tail->graphs[g1])+1 == getSize(inpCert->tail->graphs[g2]))
			{
				//This pair are potentially AI or DC1 invertable
				//Determine if their orders are compatible for an inverse op, and which one, if either, is larger.
				larger = inpCert->tail->graphs[g2];
				smaller = inpCert->tail->graphs[g1];
				
				if(inpCert->tail->signs[g1] == inpCert->tail->signs[g2])
				{
					tryAI = true;
				}
				//else if(inpCert->tail->signs[g1] != inpCert->tail->signs[g2])
				else{
					tryDC1 = true;
				}
			}
			else if(getSize(inpCert->tail->graphs[g1]) == getSize(inpCert->tail->graphs[g2]))
			{
				//This pair are potentially DC2 reversable
				//Determine if their edge counts are compatible for a DC2 inverse op, and which, if either, is larger
				if(inpCert->tail->signs[g1] != inpCert->tail->signs[g2])
				{
					if(getNumEdges(inpCert->tail->graphs[g1]) == getNumEdges(inpCert->tail->graphs[g2])+1)
					{
						larger = inpCert->tail->graphs[g1];
						smaller = inpCert->tail->graphs[g2];
						tryDC2 = true;
					}
					else if(getNumEdges(inpCert->tail->graphs[g1])+1 == getNumEdges(inpCert->tail->graphs[g2]))
					{
						larger = inpCert->tail->graphs[g2];
						smaller = inpCert->tail->graphs[g1];
						tryDC2 = true;
					}
				}			
			}
	
			//If the graph orders and signs are ok so far...
			if(tryAI || tryDC1 || tryDC2)
			{
				//Try the selected inversion for all edges
				for(i = 0; i < getSize(larger); i++)
				{
					for(j = 0; j < i; j++)
					{
						//if they have the same sign ad there is an edge at i,j
						if(tryAI && isEdge(larger,i,j))
						{
							//Contract on i and j, then compare to the smaller graph
							temp = copyAndContractEdge(larger,i,j);
							if(isomorphic(temp,smaller))
							{
								//They are reversable, so add a new operation 
								addOpToList(list, newOperation(inpCert, INVADDIDENT, g1, g2, i, j));
								//printf("%d,%d GRAPHS: %d & %d INVADDIDENT\n",i,j,getID(inpCert->tail->graphs[g1]),getID(inpCert->tail->graphs[g2]));
							}
							freeGraph(temp);
						}
						//If they have opposing signs and there is no edge at i,j
						if(tryDC1 && !isEdge(larger,i,j))
						{
							
							//Contract on i and j, then compare to the smaller graph
							temp = copyAndContractEdge(larger,i,j);
							if(isomorphic(temp,smaller))
							{
								//They are reversable, so add a new operation 
								addOpToList(list, newOperation(inpCert, INVDELCONT1, g1, g2, i, j));
								//printf("%d,%d GRAPHS: %d & %d INVDELCONT1\n",i,j,getID(inpCert->tail->graphs[g1]),getID(inpCert->tail->graphs[g2]));
							}
							freeGraph(temp);
						}
						if(tryDC2 && !isEdge(smaller,i,j))
						{
							temp = copyAndAddEdge(smaller,i,j);
							if(isomorphic(larger,temp))
							{
								//They are reversable, so add a new operation 
								addOpToList(list, newOperation(inpCert, INVDELCONT2, g1, g2, i, j));
								//printf("%d,%d GRAPHS: %d & %d INVDELCONT2\n",i,j,getID(inpCert->tail->graphs[g1]),getID(inpCert->tail->graphs[g2]));
							}
							freeGraph(temp);
						}
					}
				}
			}
			tryAI = false;
			tryDC1 = false;
			tryDC2 = false;
		}

	}
	
	return list;
}

operationList getDelContOperationList(cert inpCert)
{	
	operationList list = newOperationList();
	graph currentGraph = NULL;
	int graphSize = 0;

	int i = 0, j = 0, k = 0;
	
	for(i = 0; i < inpCert->tail->numGraphs; i++)
	{
			
		currentGraph = inpCert->tail->graphs[i];
		graphSize = getSize(currentGraph);
		if(defaultEdgePrecedence == STARTSMALL)
		{
			for(j = 0; j < graphSize; j++)
			{
				for(k = 0; k < j; k++)
				{
					if(isEdge(currentGraph,j,k))
					{
						addOpToList(list, newOperation (inpCert, DELCONT, i,0,j,k));
					}
				}
			}
		}
		else if(defaultEdgePrecedence == STARTBIG)
		{
			for(j = graphSize-2 ; j >= 0; j--)
			{
				
				for(k = graphSize-1; k >j; k--)
				{
					if(isEdge(currentGraph,j,k))
					{
						addOpToList(list, newOperation (inpCert, DELCONT, i,0,j,k));
					}
				}
			}	
		}
			
	}
	//exit(0);
	return list;
}



operationList getAddIdentOperationList(cert inpCert)
{	
	operationList list = newOperationList();
	graph currentGraph = NULL;
	int graphSize = 0;

	int i = 0, j = 0, k = 0 ;
	
	for(i = 0; i < inpCert->tail->numGraphs; i++)
	{
		currentGraph = inpCert->tail->graphs[i];
		graphSize = getSize(currentGraph);
		
		
		if(defaultEdgePrecedence == STARTSMALL)
		{
			for(j = 0; j < graphSize; j++)
			{
				for(k = 0; k < j; k++)
				{
					if(!isEdge(currentGraph,j,k))
					{
						addOpToList(list, newOperation (inpCert, ADDIDENT, i,0,j,k));
					}
				}
			}
		}
		else if(defaultEdgePrecedence == STARTBIG)
		{
			for(j = graphSize-2 ; j >= 0; j--)
			{
				
				for(k = graphSize-1; k > j; k--)
				{
					if(!isEdge(currentGraph,j,k))
					{
						addOpToList(list, newOperation (inpCert, ADDIDENT, i,0,j,k));
					}
				}
			}	
		}
	}
	
	return list;
}


operationList getEdgeOperationList(cert inpCert)
{	
	operationList list = newOperationList();
	graph currentGraph = NULL;
	int graphSize = 0;

	int i = 0, j = 0, k = 0 ;
	
	for(i = 0; i < inpCert->tail->numGraphs; i++)
	{
		currentGraph = inpCert->tail->graphs[i];
		graphSize = getSize(currentGraph);
		
		
		if(defaultEdgePrecedence == STARTSMALL)
		{
			for(j = 0; j < graphSize; j++)
			{
				for(k = 0; k < j; k++)
				{
					if(isEdge(currentGraph,j,k))
					{
						addOpToList(list, newOperation (inpCert, DELCONT, i,0,j,k));
					}
					else
					{
						addOpToList(list, newOperation (inpCert, ADDIDENT, i,0,j,k));
					}
				}
			}
		}
		else if(defaultEdgePrecedence == STARTBIG)
		{
			for(j = graphSize-2 ; j >= 0; j--)
			{
				
				for(k = graphSize-1; k > j; k--)
				{
					if(isEdge(currentGraph,j,k))
					{
						addOpToList(list, newOperation (inpCert, DELCONT, i,0,j,k));
					}
					else
					{
						addOpToList(list, newOperation (inpCert, ADDIDENT, i,0,j,k));
					}
				}
			}	
		}
		//printf("N choose 2: %d  Op count%d\n",,list->length);
	}

	return list;
}


operationList getOperationList(cert inpCert, enum operationType optype)
{
	operationList list;
	
	switch(optype)
	{
		case DELCONT:		list = getDelContOperationList(inpCert);	break;
		case ADDIDENT:		list = getAddIdentOperationList(inpCert);	break;
		case INVDELCONT1:	list = getDC1RevOperationList(inpCert);		break;
		case INVDELCONT2:	list = getDC2RevOperationList(inpCert);		break;
		case INVADDIDENT:	list = getAIRevOperationList(inpCert);		break;
		case CANCEL:		list = getCancelOperationList(inpCert);		break;
		default:		return NULL;					break;
	}
	return list;
}

bool listIsEmpty(operationList list)
{
	if(list->head == NULL)
	{
		return true;
	}
	return false;
}

int getListLength(operationList list)
{
	return list->length;
}


operationList getAllOperations(cert inpCert)
{
	operationList list = NULL, temp = NULL;
	int numTypes = 6;
	enum operationType *opTypeOrder = NULL;
	
	switch(defaultPrecedence)
	{
		case DELADDINV:
			opTypeOrder = (enum operationType []){DELCONT, ADDIDENT, INVDELCONT1, INVDELCONT2, INVADDIDENT, CANCEL};
			list = buildOperationList(inpCert, opTypeOrder, numTypes);
			break;	

		case DELOPT:
			opTypeOrder = (enum operationType []){DELCONT, INVDELCONT1, INVDELCONT2, ADDIDENT, INVADDIDENT, CANCEL};
			list = buildOperationList(inpCert, opTypeOrder, numTypes);
			break;	

		case ADDDELINV:
			opTypeOrder = (enum operationType []){ADDIDENT, DELCONT, INVADDIDENT, INVDELCONT1, INVDELCONT2, CANCEL};
			list = buildOperationList(inpCert, opTypeOrder, numTypes);
			break;	

		case ADDOPT:
			opTypeOrder = (enum operationType []) {ADDIDENT, INVADDIDENT, DELCONT, INVDELCONT1, INVDELCONT2, CANCEL};
			list = buildOperationList(inpCert, opTypeOrder, numTypes);
			break;
		case FAST:
			list = getEdgeOperationList(inpCert);
			temp = getPairwiseOperationList(inpCert);
			appendList(list,temp);
			free(temp);
			break;

		default:	
			list = NULL;
	}
	return list;
}

operationList getAllCompressingOperations(cert inpCert)
{
	operationList list = NULL, temp = NULL;
	int numTypes = 4;
	enum operationType *opTypeOrder = NULL;
	
	switch(defaultPrecedence)
	{
		case DELADDINV:		
			opTypeOrder = (enum operationType []){INVDELCONT1, INVDELCONT2, INVADDIDENT, CANCEL};;
			list = buildOperationList(inpCert, opTypeOrder, numTypes);
			break;
					
		case DELOPT:
			opTypeOrder = (enum operationType []){INVDELCONT1, INVDELCONT2, INVADDIDENT, CANCEL};;
			list = buildOperationList(inpCert, opTypeOrder, numTypes);
			break;
					
		case ADDDELINV:	
			opTypeOrder = (enum operationType []){INVADDIDENT, INVDELCONT1, INVDELCONT2, CANCEL};;
			list = buildOperationList(inpCert, opTypeOrder, numTypes);
			break;
			
		case ADDOPT:	
			opTypeOrder = (enum operationType []){INVADDIDENT, INVDELCONT1, INVDELCONT2, CANCEL};;
			list = buildOperationList(inpCert, opTypeOrder, numTypes);
			break;
		case FAST:
			list = getPairwiseOperationList(inpCert);
			break;
		default:		list = NULL;
	}
	return list;
}

operationList buildOperationList(cert inpCert, enum operationType * types, int size)
{
	operationList list = newOperationList(), temp = NULL;
	int i = 0;
	while(i<size)
	{
		temp = getOperationList(inpCert,types[i]);
		appendList(list,temp);
		free(temp);
		i++;
	}
	return list;
}
/*
operationList getAllOperations(cert inpCert)
{
	operationList list = NULL, temp = NULL;

	list = getAddIdentOperationList(inpCert);
	temp = getDelContOperationList(inpCert);
	appendList(list, temp);
	free(temp);
	temp = getAIRevOperationList(inpCert);
	appendList(list, temp);
	free(temp);
	temp = getDC1RevOperationList(inpCert);
	appendList(list, temp);
	free(temp);
	temp = getDC2RevOperationList(inpCert);
	appendList(list, temp);
	free(temp);
	temp = getCancelOperationList(inpCert);
	appendList(list, temp);
	free(temp);
	
	return list;
}

operationList getAllCompressingOperations(cert inpCert)
{
	operationList list = NULL, temp = NULL;

	list = getAIRevOperationList(inpCert);
	temp = getDC1RevOperationList(inpCert);
	appendList(list, temp);
	free(temp);
	temp = getDC2RevOperationList(inpCert);
	appendList(list, temp);
	free(temp);
	temp = getCancelOperationList(inpCert);
	appendList(list, temp);
	free(temp);
	return list;
}*/

//
/*operationList getAllOperations(cert inpCert)
{
	operationList list = NULL, temp = NULL;

	list = getDelContOperationList(inpCert);
	temp = getAddIdentOperationList(inpCert);
	appendList(list, temp);
	free(temp);
	temp = getDC1RevOperationList(inpCert);
	appendList(list, temp);
	free(temp);
	temp = getDC2RevOperationList(inpCert);
	appendList(list, temp);
	free(temp);
	temp = getAIRevOperationList(inpCert);
	appendList(list, temp);
	free(temp);
	temp = getCancelOperationList(inpCert);
	appendList(list, temp);
	free(temp);
	
	return list;
}

operationList getAllCompressingOperations(cert inpCert)
{
	operationList list = NULL, temp = NULL;

	list = getDC1RevOperationList(inpCert);
	temp = getDC2RevOperationList(inpCert);
	appendList(list, temp);
	free(temp);
	temp = getAIRevOperationList(inpCert);
	appendList(list, temp);
	free(temp);
	temp = getCancelOperationList(inpCert);
	appendList(list, temp);
	free(temp);
	return list;
}*/

/*operationList getAllOperations(cert inpCert)
{
	operationList list = NULL, temp = NULL;

	list = getDelContOperationList(inpCert);
	temp = getDC1RevOperationList(inpCert);
	appendList(list, temp);
	free(temp);
	temp = getDC2RevOperationList(inpCert);
	appendList(list, temp);
	free(temp);
	temp = getAddIdentOperationList(inpCert);
	appendList(list, temp);
	free(temp);
	temp = getAIRevOperationList(inpCert);
	appendList(list, temp);
	free(temp);
	temp = getCancelOperationList(inpCert);
	appendList(list, temp);
	free(temp);
	
	return list;
}

operationList getAllCompressingOperations(cert inpCert)
{
	operationList list = NULL, temp = NULL;

	list = getDC1RevOperationList(inpCert);
	temp = getDC2RevOperationList(inpCert);
	appendList(list, temp);
	free(temp);
	temp = getAIRevOperationList(inpCert);
	appendList(list, temp);
	free(temp);
	temp = getCancelOperationList(inpCert);
	appendList(list, temp);
	free(temp);
	return list;
}*/

//Modifies list1 so that list2 is now appended to it. Don't forget to free list2 after calling this one!
void appendList(operationList list1, operationList list2)
{

	if(listIsEmpty(list2))
	{
		//Do nothing to list1
	}
	else if(listIsEmpty(list1))
	{
		list1->head = list2->head;
		list1->tail = list2->tail;
		list1->length = list2->length;
	}
	else
	{
		list1->tail->next = list2->head;
		list1->tail = list2->tail;
		list1->length += list2->length;
	}
}

operation getOpFromList(operationList list, int opIndex)
{
	int i = 0;
	operationListNode currentNode = list->head;
	while(i != opIndex)
	{
		currentNode = currentNode->next;
		i++;
	}
	return currentNode->op;
}

void printOperationList(operationList list)
{
	int i = 0;
	operationListNode currentNode;
	
	if(listIsEmpty(list))
	{
		printf("\nThe list is empty\n");
	}
	else
	{
		operationListNode currentNode = list->head;
		printf("\nThere are %d operations: ",list->length);
		for(i = 0; i < list->length; i++)
		{
			printf("\n\t%d: ",i);
			printOperation(currentNode->op);
			currentNode = currentNode->next;
		}
		printf("\n");
	}
}





/*****************************************************************************/
/********************************TEST CODE************************************/
/*****************************************************************************/


/******************************************************************************
Function:	runComp
Description:	Tests if the graphs
Inputs:		line - the line containing the graphs
Return Value:	None
Side Effects:	None
Notes:		
Created:	Z.Bukovac Apr 2012
******************************************************************************/
void runComp(cert inpCert)
{	
	int i = 0,j = 0;
	graph g1, g2;
	line inpLine = inpCert->tail;
	//For each graph in the line
	for(i = 0; i < inpLine->numGraphs; i++)
	{
		for(j = 0; j < i; j++)
		{
			g1 = inpLine->graphs[i];
			g2 = inpLine->graphs[j];
			
			if(getSize(g1) == getSize(g2)  &&  isomorphic(g1, g2))
			{
				//DO STUFF;
			}
		}
	}
}

void certTest()
{
	graph g1 = createGraph(10);
	cert inpCert = newCert(g1);
	setID(g1,0);
	printCert(inpCert);
	deleteCont(inpCert,0,1,0);
	printCert(inpCert);
	runComp(inpCert);
	deleteCont(inpCert,1,4,1);
	printCert(inpCert);
	runComp(inpCert);
	deleteCont(inpCert,2,0,2);
	printCert(inpCert);
	runComp(inpCert);
	printCertGraphs(inpCert);
	addIdent(inpCert,1,0,8);
	printCert(inpCert);
	printCertGraphs(inpCert);
	runComp(inpCert);
	addIdent(inpCert,3,0,7);
	printCert(inpCert);
	printCancelableGraphPairs(inpCert);
	printCertGraphs(inpCert);
	runComp(inpCert);
	cancelGraphs(inpCert,1,4);
	printCert(inpCert);
	runComp(inpCert);
	cancelGraphs(inpCert,1,3);
	printCert(inpCert);
	runComp(inpCert);
	printCancelableGraphPairs(inpCert);
	runComp(inpCert);
	/*addIdent(inpCert,0,6,2);
	printCert(inpCert);
	printCertGraphs(inpCert);
	addIdent(inpCert,1,8,3);
	printCert(inpCert);
	printCertGraphs(inpCert);*/
}

void newTest()
{
	graph g1 = initGraph(5);
	addEdge(g1,0,1);
	addEdge(g1,1,2);
	addEdge(g1,2,3);
	addEdge(g1,3,4);
	
	printGraph(g1);
	printf("\nThe following four graphs should be isomorphic");
	printGraph(copyAndContractEdge(g1,0,1));
	printGraph(copyAndContractEdge(g1,1,2));
	printGraph(copyAndContractEdge(g1,2,3));
	printGraph(copyAndContractEdge(g1,3,4));
	
	addEdge(g1,0,2);
	
	printf("\nAdd (0,2) to G0:");
	printGraph(g1);
	printf("\nThe following two graphs should be isomorphic:");
	printGraph(copyAndContractEdge(g1,0,1));
	printGraph(copyAndContractEdge(g1,1,2));
	printf("\nThe following two graphs should be isomorphic:");
	printGraph(copyAndContractEdge(g1,2,3));
	printGraph(copyAndContractEdge(g1,3,4));
	
	graph g2 = initGraph(5); setID(g2,2);
	graph g3 = initGraph(4); setID(g3,3);
	
	addEdge(g2,0,2);
	addEdge(g2,1,2);
	addEdge(g2,3,2);
	addEdge(g2,4,2);
	addEdge(g2,0,1);
	
	addEdge(g3,0,1);
	addEdge(g3,0,2);
	addEdge(g3,0,3);
	
	printf("\nThe following should be isomorphic when contracting on G2(0,1)");
	printGraph(g2);
	printGraph(g3);
	
	//printf("\nContracting... Iso?:%d",isomorphic(g3,copyAndContractEdge(g2,0,1)));
	printGraph(copyAndContractEdge(g2,0,1));
	
}
/**********************************************************************************/
/**********************************EXTRAS******************************************/
/**********************************************************************************/


/******************************************************************************
Function:	getDC1Reversals
Description:	Adds new operations to the given list for any pairs of graphs that
		can be found in the inpCert->tail that will DC1 reverse.
Inputs:		list - the list to add these new operations to
		inpCert - the certificate to look at (specifically, its tail line)
		graphIndex1, graphIndex2: the indices of the graphs in question
Return Value:	None
Side Effects:	Alters list, adds all DC1 reversable graph pair ops in the cert->tail.
Notes:		
Created:	Z.Bukovac Feb 2012
******************************************************************************/
/*
void getDC1Reversals(operationList list, cert inpCert, int graphIndex1, int graphIndex2)
{
	//The smaller graph needs to be negative
	graph g1 = inpCert->tail->graphs[graphIndex1];
	graph g2 = inpCert->tail->graphs[graphIndex2];
	
	operation op;
	graph larger, smaller;
	int largerIndex, smallerIndex;		//The index of the larger graph
	int i,j = 0;

	//Determine if their orders are compatible, and which one, if either, is larger.
	if(getSize(g1) == getSize(g2)+1)
	{
		larger = g1;
		smaller = g2;
		largerIndex = graphIndex1;
		smallerIndex = graphIndex2;
	}
	else if(getSize(g1)+1 == getSize(g2))
	{
		larger = g2;
		smaller = g1;
		largerIndex = graphIndex2;
		smallerIndex = graphIndex1;
	}
	else
	{
		//Their orders are incompatible for a reverse DC1 op.
		return;	//return NULL;
	}
	
	//If they have opposing signs...
	if(inpCert->tail->signs[graphIndex1] != inpCert->tail->signs[graphIndex2])
	{
		//Search through the non-edges that when contracted in the larger graph produce the smaller graph.
		//For each non-edge in the graph...
			//Try contracting on its vertices and then check if (contract(larger) == smaller)
			//If so, it's reversable, so return a new operation stuct with the relevant details.

		for(i = 0; i < getSize(larger); i++)
		{
			for(j = 0; j < i; j++)
			{
				//If there is no edge at ij...
				if(!isEdge(larger,i,j))
				{
					//Contract on i and j, then compare to the smaller graph
					if(isomorphic(copyAndContractEdge(larger,i,j),smaller))
					{
						//TODO They are reversable, so add a new operation 
						addOpToList(list, newOperation(inpCert, INVDELCONT1, largerIndex, smallerIndex, i, j));
					}
				}
			}
		}
	}
}
*/


